import pathlib
import pickle
import os

def initialize_pywit_directory() -> None:
    home_path = pathlib.Path.home()
    paths = [home_path.joinpath('pywit').joinpath(ext) for ext in ('config', 'IW2D/bin', 'IW2D/projects')]
    for path in paths:
        os.makedirs(path, exist_ok=True)
    with open(pathlib.Path(paths[0]).joinpath('iw2d_settings.yaml'), 'w') as file:
        file.write(f'binary_directory: {paths[1]}\n'
                   f'project_directory: {paths[2]}')

    filenames = ('component', 'element', 'iw2d_inputs')
    for filename in filenames:
        open(paths[0].joinpath(f"{filename}.yaml"), 'w').close()

    with open(paths[2].joinpath('hashmap.pickle'), 'wb') as handle:
        pickle.dump(dict(), handle, protocol=pickle.HIGHEST_PROTOCOL)



from typing import Tuple,Sequence
from .component import Component
from .element import Element
from .utilities import create_resonator_component
from .interface import component_names, get_component_name

import numpy as np
from scipy import integrate

from scipy.constants import c as c_light, mu_0
from scipy.special import erfc, zeta

free_space_impedance = mu_0 * c_light


def create_tesla_cavity_component(plane: str, exponents: Tuple[int, int, int, int], a: float, g: float,
                                  period_length: float) -> Component:
    """
    Creates a single component object modeling a periodic accelerating stucture.
    Follow K. Bane formalism developped in SLAC-PUB-9663, "Short-range dipole wakefields
    in accelerating structures for the NLC".
    :param plane: the plane the component corresponds to
    :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
    :param a: accelerating structure iris gap in m
    :param g: individual cell gap in m
    :param period_length: period length in m
    :return: A component object of a periodic accelerating structure
    """

    # Material properties required for the skin depth computation are derived from the input Layer attributes
    # material_resistivity = layer.dc_resistivity
    # material_relative_permeability = layer.magnetic_susceptibility
    # material_permeability = material_relative_permeability * scipy.constants.mu_0

    # Create the skin depth as a function of frequency and layer properties
    # delta_skin = lambda freq: (material_resistivity/ (2*pi*abs(freq) * material_permeability)) ** (1/2)

    gamma = g / period_length
    alpha1 = 0.4648
    alpha = 1 - alpha1 * np.sqrt(gamma) - (1 - 2*alpha1) * gamma

    s00 = g / 8 * (a / (alpha * period_length)) ** 2

    # Longitudinal impedance and wake
    if plane == 'z' and exponents == (0, 0, 0, 0):
        def longitudinal_impedance_tesla_cavity(freq):
            return (1j * free_space_impedance / (np.pi * (2 * np.pi * freq / c_light) * a ** 2) *
                    (1 + (1 + 1j) * alpha * period_length / a * np.sqrt(np.pi / ((2 * np.pi * freq / c_light) * g))) ** (-1))

        def longitudinal_wake_tesla_cavity(time):
            return ((free_space_impedance * c_light / (np.pi * a ** 2)) * np.heaviside(time, 0) *
                    np.exp(-np.pi * c_light * time / (4 * s00)) * erfc(np.sqrt(np.pi * c_light * time / (4 * s00))))

        impedance = longitudinal_impedance_tesla_cavity
        wake = longitudinal_wake_tesla_cavity
    # Transverse dipolar impedance and wake
    elif (plane == 'x' and exponents == (1, 0, 0, 0)) or (plane == 'y' and exponents == (0, 1, 0, 0)):

        def transverse_dipolar_impedance_tesla_cavity(freq):
            return 2 / ((2 * np.pi * freq / c_light) * a ** 2) * 1j * free_space_impedance / (
                    np.pi * (2 * np.pi * freq / c_light) * a ** 2) * (
                    1 + (1 + 1j) * alpha * period_length / a * np.sqrt(np.pi / ((2 * np.pi * freq / c_light) * g))) ** (-1)

        def transverse_dipolar_wake_tesla_cavity(time):
            return ((4 * free_space_impedance * c_light * s00) / (np.pi * a ** 4) * np.heaviside(time, 0) *
                    (1 - (1 + np.sqrt(c_light * time / s00)) * np.exp(-np.sqrt(c_light * time / s00))))

        impedance = transverse_dipolar_impedance_tesla_cavity
        wake = transverse_dipolar_wake_tesla_cavity
    else:
        print("Warning: tesla cavity impedance not implemented for component {}{}. Set to zero".format(plane,
                                                                                                       exponents))

        def zero_function(x):
            return np.zeros_like(x)

        impedance = zero_function
        wake = zero_function

    return Component(impedance=np.vectorize(lambda f: impedance(f)),
                     wake=np.vectorize(lambda t: wake(t)),
                     plane=plane, source_exponents=exponents[:2], test_exponents=exponents[2:])


def _integral_stupakov(half_gap_small: float, half_gap_big: float,
                       half_width: float, g_index: int, g_power: int,
                       approximate_integrals: bool = False):
    """
    Computes the Stupakov's integral for a rectangular linear taper.
    Note that (g')^2 has been taken out of the integral. See Phys. Rev. STAB 10, 094401 (2007)
    :param half_gap_small: the small half-gap of the taper
    :param half_gap_big: the large half-gap of the taper
    :param half_width: the half-width of the taper
    :param g_index: the indice of the G function to use (0 is for G0=F in Stupakov's paper)
    :param g_power: is the power to which 1/g is taken
    :param approximate_integrals: use approximated formulas to compute the integrals.
    It can be used if one assumes small half_gap_big/half_width ratio << 1
    """
    def _integrand_stupakov(g):
        """
        Computes the integrand for the Stupakov integral
        :param g: the half-gap of the taper
        """
        x = g/half_width # half-gap over half-width ratio

        def g_addend(m):
            """
            Computes the m-th addend of series defining F (when g_index=0) or G_[g_index] function from Stupakov's formulas for a
            rectangular linear taper at a given x=half-gap/half-width. m can be an array, in which case the function returns an array.
            See Phys. Rev. STAB 10, 094401 (2007)
            :param m: the index of the addend (it can be an array)
            """

            if g_index == 0:
                val = (2 * m + 1) * np.pi * x / 2
                res = ((1 - np.exp(-2 * val)) / (1 + np.exp(-2 * val)) * (
                        2 * (np.exp(-val)) / (1 + np.exp(-2 * val))) ** 2) / (2 * m + 1)

                return res

            elif g_index == 1:
                val = (2 * m + 1) * np.pi * x / 2
                res = x ** 3 * (2 * m + 1) * (4 * np.exp(-2 * val)) * (1 + np.exp(-2 * val)) / ((1 - np.exp(-2 * val)) ** 3)
                return res

            elif g_index == 2:
                val = (2 * m + 1) * np.pi * x / 2
                res = x ** 2 * (2 * m + 1) * (
                    ((1 - np.exp(-2 * val)) / (1 + np.exp(-2 * val)) * (2 * (np.exp(-val)) / (1 + np.exp(-2 * val))) ** 2))
                return res

            elif g_index == 3:
                val = m * np.pi * x
                res = x ** 2 * (2 * m) * ((1 - np.exp(-2 * val)) / (1 + np.exp(-2 * val)) * (
                        2 * (np.exp(-val)) / (1 + np.exp(-2 * val))) ** 2)
                return res

            else:
                raise ValueError("Pb in g_addend for Stupakov's formulas: g_index not 0, 1, 2 or 3 !")

        # Computes F (when g_index=0) or G_[g_index] function from Stupakov's formulas.
        g_stupakov = 0.
        old_g_stupakov = 1.
        eps = 1e-5  # relative precision of the summation
        incr = 10  # increment for sum computation
        m = np.arange(incr)
        m_limit = 1e6

        while (abs(g_stupakov - old_g_stupakov) > eps * abs(g_stupakov)) and (m[-1] < m_limit):
            g_array = g_addend(m)
            old_g_stupakov = g_stupakov
            g_stupakov += np.sum(g_array)
            m += incr

        if m[-1] >= m_limit:
            print("Warning: maximum number of elements reached in g_stupakov!", m[-1], x, ", err=",
                  abs((g_stupakov - old_g_stupakov) / g_stupakov))

        return g_stupakov / (g ** g_power)

    if approximate_integrals:
        if g_index == 0 and g_power == 0:
            i = 7. * zeta(3, 1) / (2. * np.pi ** 2) * (
                    half_gap_big - half_gap_small)  # (zeta(3.,1.) is Riemann zeta function at x=3)

        elif g_index == 1 and g_power == 3:
            i = (1. / (half_gap_small ** 2) - 1. / (half_gap_big ** 2)) / (
                        2. * np.pi)  # approx. integral

        elif g_index in [2,3] and g_power == 2:
            i = (1. / half_gap_small - 1. / half_gap_big) / (np.pi ** 2)

        else:
            raise ValueError("Wrong values of g_index and g_power in _integral_stupakov")

    else:
        # computes numerically the integral instead of using its approximation
        i, err = integrate.quadrature(_integrand_stupakov, half_gap_small, half_gap_big,
                                      tol=1.e-3, maxiter=200, vec_func=False)

    return i


def shunt_impedance_flat_taper_stupakov_formula(half_gap_small: float, half_gap_big: float, taper_slope: float,
                                                half_width: float, cutoff_frequency: float = None,
                                                component_id: str = None, approximate_integrals: bool = False) -> float:
    """
    Computes the shunt impedance in Ohm(/m if not longitudinal) of one single rectangular linear taper using Stupakov's
    formulae (Phys. Rev. STAB 10, 094401 - 2007), multiplied by Z0*c/(4*pi) to convert to SI units.
    Taper is in the vertical plane (i.e. the change of half-gap is in vertical, but the width is along the horizontal plane
    - this is typical of a taper going towards a vertical collimator with horizontal jaws).
    We use here half values for geometrical parameters (half-gap and half-width) whereas Stupakov's paper is expressed
    with full values. This does not make any difference except for an additional factor 4 here for longitudinal impedance.

    The formula is valid under the conditions of low frequency and length of taper much larger than its transverse
    dimensions.

    Note: one gets zero if component_id is not in [zlong, zxdip, zydip, zxqua, zyqua].

    :param half_gap_small: small vertical half-gap
    :param half_gap_big: large vertical half-gap
    :param taper_slope: the slope of the taper
    :param half_width: half width of the taper (constant)
    :param cutoff_frequency: the cutoff frequency (used only for the longitudinal component)
    :param component_id: a component name for which one computes the R/Q
    (ex: zlong, zydip, zxqua, etc.)
    :param approximate_integrals: use approximated formulas to compute the integrals. It can be used if one assumes
    small half_gap_big/half_width ratio (<<1)
    :return: the shunt impedance of the component.
    """
    z_0 = mu_0 * c_light  # free space impedance

    if cutoff_frequency is None and component_id == 'zlong':
        raise ValueError("cutoff_frequency must be specified when component_id is 'zlong'")

    if component_id == 'zlong':
        g_index = 0
        g_power = 0
        cst = 4. * mu_0 * cutoff_frequency / 2.  # factor 4 due to use of half-gaps here

    elif component_id == 'zydip':
        g_index = 1
        g_power = 3
        cst = z_0 * half_width * np.pi / 4.

    elif component_id == 'zxqua':
        g_index = 2
        g_power = 2
        cst = -z_0 * np.pi / 4.

    elif component_id == 'zxdip':
        g_index = 3
        g_power = 2
        cst = z_0 * np.pi / 4.

    elif component_id == 'zyqua':
        g_index = 2
        g_power = 2
        cst = z_0 * np.pi / 4.
    else:
        # mock-up values
        g_power = 0
        g_index = 0
        cst = 0

    # computes the integral
    i = _integral_stupakov(half_gap_small, half_gap_big, half_width, g_index, g_power,
                           approximate_integrals = approximate_integrals)
    i *= taper_slope  # put back g' factor that was dropped

    return cst * i


def create_flat_taper_stupakov_formula_component(half_gap_small: float, half_gap_big: float, taper_slope: float,
                                                 half_width: float, plane: str, exponents: Tuple[int, int, int, int],
                                                 cutoff_frequency: float = None) -> Component:
    """
    Creates a component using the flat taper Stupakov formula
    :param half_gap_small: small vertical half-gap
    :param half_gap_big: large vertical half-gap
    :param taper_slope: the slope of the taper
    :param half_width: half width of the taper (constant)
    :param plane: the plane the component corresponds to
    :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
    :param cutoff_frequency: the cutoff frequency (used only for the longitudinal component)
    :return: A component object of a flat taper
    """
    component_id = get_component_name(True, plane, exponents)

    r_shunt = shunt_impedance_flat_taper_stupakov_formula(half_gap_small=half_gap_small, half_gap_big=half_gap_big,
                                                          taper_slope=taper_slope, half_width=half_width,
                                                          cutoff_frequency=cutoff_frequency, component_id=component_id)

    return create_resonator_component(plane=plane, exponents=exponents, r=r_shunt, q=1, f_r=cutoff_frequency)


def create_flat_taper_stupakov_formula_element(half_gap_small: float, half_gap_big: float, taper_slope: float,
                                               half_width: float, beta_x: float, beta_y: float,
                                               cutoff_frequency: float = None,
                                               component_ids: Sequence[str] = ('zlong', 'zxdip', 'zydip', 'zxqua', 'zyqua'),
                                               name: str = "Flat taper", tag: str = "",
                                               description: str = "") -> Element:
    """
    Creates an element using the flat taper Stupakov formula
    :param half_gap_small: small vertical half-gap
    :param half_gap_big: large vertical half-gap
    :param taper_slope: the slope of the taper
    :param half_width: half width of the taper (constant)
    :param beta_x: The beta function in the x-plane at the position of the taper
    :param beta_y: The beta function in the y-plane at the position of the taper
    :param cutoff_frequency: the cutoff frequency (used only for the longitudinal component)
    :param component_ids: a list of components to be computed
    :param name: A user-specified name for the Element
    :param tag: A string to tag the Element
    :param description: A description for the Element
    :return: An Element object representing the flat taper
    """
    length = (half_gap_big - half_gap_small) / taper_slope

    components = []
    for component_id in component_ids:
        _, plane, exponents = component_names[component_id]
        components.append(create_flat_taper_stupakov_formula_component(half_gap_small=half_gap_small,
                                                                       half_gap_big=half_gap_big,
                                                                       taper_slope=taper_slope,
                                                                       half_width=half_width,
                                                                       plane=plane, exponents=exponents,
                                                                       cutoff_frequency=cutoff_frequency,
                                                                       ))

    return Element(length=length, beta_x=beta_x, beta_y=beta_y, components=components, name=name, tag=tag,
                   description=description)



from .component import Component
from .element import Element
from .interface import Layer, FlatIW2DInput, RoundIW2DInput
from .interface import component_names

from yaml import load, SafeLoader
from typing import Tuple, Dict, List, Union, Sequence, Optional, Callable
from collections import defaultdict

from numpy import (vectorize, sqrt, exp, pi, sin, cos, abs, sign,
                   inf, floor, linspace, ones, isscalar, array)
from numpy.typing import ArrayLike
import scipy.constants
from scipy import special as sp
import numpy as np

if hasattr(np, 'trapezoid'):
    trapz = np.trapezoid # numpy 2.0
else:
    trapz = np.trapz

c_light = scipy.constants.speed_of_light  # m s-1
mu0 = scipy.constants.mu_0
Z0 = mu0 * c_light
eps0 = scipy.constants.epsilon_0


def string_to_params(name: str, include_is_impedance: bool = True):
    """
    Converts a string describing some specific component to the set of parameters which give this component
    :param name: The string description of the component to be converted. Either of the form <pabcd> where p is either
    'x', 'y' or 'z', and abcd are four integers giving the source- and test exponents of the component. Optionally,
    a single character, 'z' or 'w' can be included at the front of the string if it is necessary to indicate that the
    given component is either a wake- or impedance component. In this case, the include_is_impedance flag must be set.
    :param include_is_impedance: A flag indicating whether or not "name" includes a 'z' or 'w' at the beginning.
    :return: A tuple containing the necessary information to describe the type of the given component
    """
    is_impedance = False
    if include_is_impedance:
        is_impedance = name[0] == 'z'
        name = name[1:]

    plane = name[0].lower()
    exponents = tuple(int(n) for n in name[1:])

    return (is_impedance, plane, exponents) if include_is_impedance else (plane, exponents)


def create_component_from_config(identifier: str) -> Component:
    """
    Creates a Component object as specified by a .yaml config file
    :param identifier: A unique identifier-string corresponding to a component specification in config/component.yaml
    :return: A Component object initialized according to the specification
    """
    with open("config/component.yaml", "r") as f:
        cdict = load(f, Loader=SafeLoader)[identifier]
        wake = vectorize(lambda x: eval(cdict['wake'], {'exp': exp, 'sqrt': sqrt,
                                                        'pi': pi, 'x': x})) if 'wake' in cdict else None
        impedance = vectorize(lambda x: eval(cdict['impedance'], {'exp': exp, 'sqrt': sqrt,
                                                                  'pi': pi, 'x': x})) if 'impedance' in cdict else None
        name = cdict['name'] if 'name' in cdict else ""
        plane = cdict['plane']
        source_exponents = (int(cdict['source_exponents'][0]), int(cdict['source_exponents'][-1]))
        test_exponents = (int(cdict['test_exponents'][0]), int(cdict['test_exponents'][-1]))
        return Component(impedance, wake, plane, source_exponents, test_exponents, name)


def create_element_from_config(identifier: str) -> Element:
    """
    Creates an Element object as specified by a .yaml config file
    :param identifier: A unique identifier-string corresponding to an element specification in config/element.yaml
    :return: An Element object initialized according to the specifications
    """
    with open("config/element.yaml", "r") as f:
        edict = load(f, Loader=SafeLoader)[identifier]
        length = float(edict['length']) if 'length' in edict else 0
        beta_x = float(edict['beta_x']) if 'beta_x' in edict else 0
        beta_y = float(edict['beta_y']) if 'beta_y' in edict else 0
        name = edict['name'] if 'name' in edict else ""
        tag = edict['tag'] if 'tag' in edict else ""
        components = [create_component_from_config(c_id) for c_id in edict['components'].split()] \
            if 'components' in edict else []
        return Element(length, beta_x, beta_y, components, name, tag)


def compute_resonator_f_roi_half_width(q: float, f_r: float, f_roi_level: float = 0.5):
    aux = np.sqrt((1 - f_roi_level) / f_roi_level)

    return (aux + np.sqrt(aux**2 + 4*q**2))*f_r/(2*q) - f_r


def create_resonator_component(plane: str, exponents: Tuple[int, int, int, int],
                               r: float, q: float, f_r: float, f_roi_level: float = 0.5) -> Component:
    """
    Creates a single component object belonging to a resonator
    :param plane: the plane the component corresponds to
    :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
    :param r: the shunt impedance of the given component of the resonator
    :param q: the quality factor of the given component of the resonator
    :param f_r: the resonance frequency of the given component of the resonator
    :param f_roi_level: fraction of the peak ok the resonator which is covered by the ROI. I.e. the roi will cover
    the frequencies for which the resonator impedance is larger than f_roi_level*r
    :return: A component object of a resonator, specified by the input arguments
    """
    root_term = sqrt(1 - 1 / (4 * q ** 2) + 0J)
    omega_r = 2 * pi * f_r
    if plane == 'z':
        impedance = lambda f: r / (1 - 1j * q * (f_r / f - f / f_r))
        omega_bar = omega_r * root_term
        alpha = omega_r / (2 * q)
        wake = lambda t: (omega_r * r * exp(-alpha * t) * (
                    cos(omega_bar * t) - alpha * sin(omega_bar * t) / omega_bar) / q).real
    else:
        impedance = lambda f: (f_r * r) / (f * (1 - 1j * q * (f_r / f - f / f_r)))
        wake = lambda t: (omega_r * r * exp(-omega_r * t / (2 * q)) * sin(omega_r * root_term * t) /
                          (q * root_term)).real

    if q > 1:
        d = compute_resonator_f_roi_half_width(q=q, f_r=f_r, f_roi_level=f_roi_level)
        f_rois = [(f_r - d, f_r + d)]
    else:
        f_rois = []
    # TODO: add ROI(s) for wake

    return Component(impedance, wake, plane, source_exponents=exponents[:2],
                     test_exponents=exponents[2:],
                     f_rois=f_rois)


def create_resonator_element(length: float, beta_x: float, beta_y: float,
                             rs: Dict[str, float], qs: Dict[str, float], fs: Dict[str, float],
                             f_roi_levels: Dict[str, float] = None, tag: str = 'resonator',
                             description: str = '') -> Element:
    """
    Creates an element object representing a resonator.
    :param length: The length, in meters, of the resonator element
    :param beta_x: The value of the beta function in the x-direction at the position of the resonator element
    :param beta_y: The value of the beta function in the y-direction at the position of the resonator element
    :param rs: A dictionary where the keys correspond to a plane followed by four exponents, i.e. "y0100", and the
    values give the Shunt impedance corresponding to this particular component
    :param qs: A dictionary where the keys correspond to a plane followed by four exponents, i.e. "y0100", and the
    values give the quality factor of the specified component of the resonator
    :param fs: A dictionary where the keys correspond to a plane followed by four exponents, i.e. "y0100", and the
    values give the resonance frequency corresponding to the particular component
    :param f_roi_levels: A dictionary where the keys correspond to a plane followed by four exponents, i.e. "y0100", and
    the values give the fraction of the peak ok the resonator which is covered by the ROI. I.e. the roi will cover
    the frequencies for which the resonator impedance is larger than f_roi_level*r
    :param tag: An optional short string used to place elements into categories
    :param description: An optional short description of the element
    :return: An element object as specified by the user-input
    """
    if f_roi_levels is None:
        f_roi_levels = {}
        for key in rs.keys():
            f_roi_levels[key] = 0.5

    assert set(rs.keys()) == set(qs.keys()) == set(fs.keys()) == set(f_roi_levels.keys()), "The three input " \
                                                                                         "dictionaries describing " \
                                                                                         "the  resonator do not all " \
                                                                                         "have identical keys"
    components = []
    for key in rs.keys():
        plane, exponents = string_to_params(key, include_is_impedance=False)
        components.append(create_resonator_component(plane, exponents, rs[key], qs[key], fs[key],
                                                     f_roi_level=f_roi_levels[key]))

    return Element(length, beta_x, beta_y, components, tag=tag, description=description)


def create_many_resonators_element(length: float, beta_x: float, beta_y: float,
                                   params_dict: Dict[str, List[Dict[str, float]]], tag: str = 'resonator',
                                   description: str = '') -> Element:
    """
    Creates an element object representing many resonators.
    :param length: The length, in meters, of the element
    :param beta_x: The value of the beta function in the x-direction at the position of the element
    :param beta_y: The value of the beta function in the y-direction at the position of the element
    :param params_dict: a dictionary associating to each component a list of dictionaries containing the
    parameters of a resonator component. E.g.:
    params_dict = {
        'z0000':
            [
                {'r': 10, 'q': 10, 'f': 50, 'f_roi_level: 0.5},
                {'r': 40, 'q': 100, 'f': 60}
            ],
        'x1000':
        ...
    }
    f_roi_level is optional
    :param tag: An optional short string used to place elements into categories
    :param description: An optional short description of the element
    :return: An element object as specified by the user-input
    """
    all_components = []
    for component_id, component_params_list in params_dict.items():
        plane, exponents = string_to_params(component_id, include_is_impedance=False)
        for component_params in component_params_list:
            assert ('r' in component_params.keys() and 'q' in component_params.keys() and
                    'f' in component_params.keys()), "each of the the component dictionaries must contain r, q and f"

            f_roi_level = component_params.get('f_roi_level', 0.5)
            all_components.append(create_resonator_component(plane, exponents, component_params['r'],
                                                             component_params['q'], component_params['f'],
                                                             f_roi_level=f_roi_level))

    comp_dict = defaultdict(lambda: 0)
    for c in all_components:
        comp_dict[(c.plane, c.source_exponents, c.test_exponents)] += c

    components = sorted(comp_dict.values(), key=lambda x: (x.plane, x.source_exponents, x.test_exponents))

    return Element(length, beta_x, beta_y, components, tag=tag, description=description)


def create_classic_thick_wall_component(plane: str, exponents: Tuple[int, int, int, int],
                                        layer: Layer, radius: float) -> Component:
    """
    Creates a single component object modeling a resistive wall impedance/wake,
    based on the "classic thick wall formula" (see e.g. A. W. Chao, chap. 2 in
    "Physics of Collective Beams Instabilities in High Energy Accelerators",
    John Wiley and Sons, 1993).
    Only longitudinal and transverse dipolar impedances are supported here.
    :param plane: the plane the component corresponds to
    :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
    :param layer: the chamber material, as a wit Layer object
    :param radius: the chamber radius in m
    :return: A component object
    """

    # Material properties required for the skin depth computation are derived from the input Layer attributes
    material_resistivity = layer.dc_resistivity
    material_relative_permeability = 1. + layer.magnetic_susceptibility
    material_permeability = material_relative_permeability * mu0

    # Create the skin depth as a function offrequency and layer properties
    delta_skin = lambda f: (material_resistivity / (2*pi*abs(f) * material_permeability)) ** (1/2)

    # Longitudinal impedance and wake
    if plane == 'z' and exponents == (0, 0, 0, 0):
        impedance = lambda f: (1/2) * (1+sign(f)*1j) * material_resistivity / (pi * radius) * (1 / delta_skin(f))
        wake = lambda t: - c_light / (2*pi*radius) * (Z0 * material_resistivity/pi)**(1/2) * 1/(t**(1/2))
    # Transverse dipolar impedance
    elif (plane == 'x' and exponents == (1, 0, 0, 0)) or (plane == 'y' and exponents == (0, 1, 0, 0)):
        impedance = lambda f: ((c_light/(2*pi*f)) * (1+sign(f)*1j) *
                               material_resistivity / (pi * radius**3) *
                               (1 / delta_skin(f)))
        wake = lambda t: -c_light / (2*pi*radius**3) * (Z0 * material_resistivity/pi)**(1/2) * 1/(t**(3/2))
    else:
        print("Warning: resistive wall impedance not implemented for component {}{}. Set to zero".format(plane,
                                                                                                         exponents))
        impedance = lambda f: 0
        wake = lambda f: 0

    return Component(vectorize(impedance), vectorize(wake), plane, source_exponents=exponents[:2],
                     test_exponents=exponents[2:])


def _zlong_round_single_layer_approx(frequencies: ArrayLike, gamma: float,
                                     layer: Layer, radius: float, length: float) -> ArrayLike:
    """
    Function to compute the longitudinal resistive-wall impedance from
    the single-layer, approximated formula for a cylindrical structure,
    by E. Metral (see e.g. Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10, TUPD053,
    https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
    Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401, 2009,
    https://doi.org/10.1103/PhysRevSTAB.12.084401)
    :param frequencies: the frequencies (array) (in Hz)
    :param gamma: relativistic mass factor
    :param layer: a layer with material properties (only resistivity,
    relaxation time and magnetic susceptibility are taken into account
    at this stage)
    :param radius: the radius of the structure (in m)
    :param length: the total length of the resistive object (in m)
    :return: Zlong, the longitudinal impedance at these frequencies
    """
    beta = sqrt(1.-1./gamma**2)
    omega = 2*pi*frequencies
    k = omega/(beta*c_light)

    rho = layer.dc_resistivity
    tau = layer.resistivity_relaxation_time
    mu1 = 1.+layer.magnetic_susceptibility
    eps1 = 1. - 1j/(eps0*rho*omega*(1.+1j*omega*tau))
    nu = k*sqrt(1.-beta**2*eps1*mu1)

    coef_long = 1j*omega*mu0*length/(2.*pi*beta**2*gamma**2)

    x1 = k*radius/gamma
    x1sq = x1**2
    x2 = nu*radius

    zlong = coef_long * (sp.k0(x1)/sp.i0(x1) - 1./(x1sq*(1./2.+eps1*sp.kve(1, x2)/(x2*sp.kve(0, x2)))))

    return zlong


def _zdip_round_single_layer_approx(frequencies: ArrayLike, gamma: float,
                                    layer: Layer, radius: float, length: float) -> ArrayLike:
    """
    Function to compute the transverse dipolar resistive-wall impedance from
    the single-layer, approximated formula for a cylindrical structure,
    Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10, TUPD053,
    https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
    Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401, 2009,
    https://doi.org/10.1103/PhysRevSTAB.12.084401)
    :param frequencies: the frequencies (array) (in Hz)
    :param gamma: relativistic mass factor
    :param layer: a layer with material properties (only resistivity,
    relaxation time and magnetic susceptibility are taken into account
    at this stage)
    :param radius: the radius of the structure (in m)
    :param length: the total length of the resistive object (in m)
    :return: Zdip, the transverse dipolar impedance at these frequencies
    """
    beta = sqrt(1.-1./gamma**2)
    omega = 2*pi*frequencies
    k = omega/(beta*c_light)

    rho = layer.dc_resistivity
    tau = layer.resistivity_relaxation_time
    mu1 = 1.+layer.magnetic_susceptibility
    eps1 = 1. - 1j/(eps0*rho*omega*(1.+1j*omega*tau))
    nu = k*sqrt(1.-beta**2*eps1*mu1)

    coef_dip = 1j*k**2*Z0*length/(4.*pi*beta*gamma**4)

    x1 = k*radius/gamma
    x1sq = x1**2
    x2 = nu*radius

    zdip = coef_dip * (sp.k1(x1)/sp.i1(x1) + 4.*beta**2*gamma**2/(x1sq*(2.+x2*sp.kve(0, x2)/(mu1*sp.kve(1, x2)))))

    return zdip


def create_resistive_wall_single_layer_approx_component(plane: str, exponents: Tuple[int, int, int, int],
                                                        input_data: Union[FlatIW2DInput, RoundIW2DInput]) -> Component:
    """
    Creates a single component object modeling a resistive wall impedance,
    based on the single-layer approximated formulas by E. Metral (see e.g.
    Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10, TUPD053,
    https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
    Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401, 2009,
    https://doi.org/10.1103/PhysRevSTAB.12.084401)
    :param plane: the plane the component corresponds to
    :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
    :param input_data: an IW2D input object (flat or round). If the input
    is of type FlatIW2DInput and symmetric, we apply to the round formula the
    Yokoya factors for an infinitely flat structure (see e.g. K. Yokoya,
    KEK Preprint 92-196 (1993), and Part. Accel. 41 (1993) pp.221-248,
    https://cds.cern.ch/record/248630/files/p221.pdf),
    while for a single plate we use those from A. Burov and V. Danilov,
    PRL 82,11 (1999), https://doi.org/10.1103/PhysRevLett.82.2286. Other
    kinds of asymmetric structure will raise an error.
    If the input is of type RoundIW2DInput, the structure is in principle round
    but the Yokoya factors put in the input will be used.
    :return: A component object
    """
    gamma = input_data.relativistic_gamma
    length = input_data.length

    if isinstance(input_data, FlatIW2DInput):
        if len(input_data.top_layers) > 1:
            raise NotImplementedError("Input data can have only one layer")
        yok_long = 1.
        layer = input_data.top_layers[0]
        radius = input_data.top_half_gap
        if input_data.top_bottom_symmetry:
            yok_dipx = pi**2/24.
            yok_dipy = pi**2/12.
            yok_quax = -pi**2/24.
            yok_quay = pi**2/24.
        elif input_data.bottom_half_gap == inf:
            yok_dipx = 0.25
            yok_dipy = 0.25
            yok_quax = -0.25
            yok_quay = 0.25
        else:
            raise NotImplementedError("For asymmetric structures, only the case of a single plate is implemented; "
                                      "hence the bottom half gap must be infinite")
    elif isinstance(input_data, RoundIW2DInput):
        radius = input_data.inner_layer_radius
        if len(input_data.layers) > 1:
            raise NotImplementedError("Input data can have only one layer")
        layer = input_data.layers[0]
        yok_long = input_data.yokoya_factors[0]
        yok_dipx = input_data.yokoya_factors[1]
        yok_dipy = input_data.yokoya_factors[2]
        yok_quax = input_data.yokoya_factors[3]
        yok_quay = input_data.yokoya_factors[4]
    else:
        raise NotImplementedError("Input of type neither FlatIW2DInput nor RoundIW2DInput cannot be handled")

    # Longitudinal impedance
    if plane == 'z' and exponents == (0, 0, 0, 0):
        impedance = lambda f: yok_long*_zlong_round_single_layer_approx(
                                            f, gamma, layer, radius, length)
    # Transverse impedances
    elif plane == 'x' and exponents == (1, 0, 0, 0):
        impedance = lambda f: yok_dipx*_zdip_round_single_layer_approx(
                                            f, gamma, layer, radius, length)
    elif plane == 'y' and exponents == (0, 1, 0, 0):
        impedance = lambda f: yok_dipy*_zdip_round_single_layer_approx(
                                            f, gamma, layer, radius, length)
    elif plane == 'x' and exponents == (0, 0, 1, 0):
        impedance = lambda f: yok_quax*_zdip_round_single_layer_approx(
                                            f, gamma, layer, radius, length)
    elif plane == 'y' and exponents == (0, 0, 0, 1):
        impedance = lambda f: yok_quay*_zdip_round_single_layer_approx(
                                            f, gamma, layer, radius, length)
    else:
        impedance = lambda f: 0

    return Component(impedance=impedance, plane=plane, source_exponents=exponents[:2],
                     test_exponents=exponents[2:])


def create_resistive_wall_single_layer_approx_element(
            input_data: Union[FlatIW2DInput, RoundIW2DInput],
            beta_x: float, beta_y: float,
            component_ids: Sequence[str] = ('zlong', 'zxdip', 'zydip', 'zxqua', 'zyqua'),
            name: str = "", tag: str = "", description: str = "") -> Element:
    """
    Creates an element object modeling a resistive wall impedance,
    based on the single-layer approximated formulas by E. Metral (see e.g.
    Eqs. 13-14 in N. Mounet and E. Metral, IPAC'10, TUPD053,
    https://accelconf.web.cern.ch/IPAC10/papers/tupd053.pdf, and
    Eq. 21 in F. Roncarolo et al, Phys. Rev. ST Accel. Beams 12, 084401, 2009,
    https://doi.org/10.1103/PhysRevSTAB.12.084401)
    :param input_data: an IW2D input object (flat or round). If the input
    is of type FlatIW2DInput and symmetric, we apply to the round formula the
    Yokoya factors for an infinitely flat structure (see e.g. K. Yokoya,
    KEK Preprint 92-196 (1993), and Part. Accel. 41 (1993) pp.221-248,
    https://cds.cern.ch/record/248630/files/p221.pdf),
    while for a single plate we use those from A. Burov and V. Danilov,
    PRL 82,11 (1999), https://doi.org/10.1103/PhysRevLett.82.2286. Other
    kinds of asymmetric structure will raise an error.
    If the input is of type RoundIW2DInput, the structure is in principle round
    but the Yokoya factors put in the input will be used.
    :param beta_x: The beta function in the x-plane at the position of the element
    :param beta_y: The beta function in the y-plane at the position of the element
    :param component_ids: a list of components to be computed
    :param name: A user-specified name for the Element
    :param tag: A string to tag the Element
    :param description: A description for the Element
    :return: An Element object representing the structure
    """
    components = []
    length = input_data.length
    for component_id in component_ids:
        _, plane, exponents = component_names[component_id]
        components.append(create_resistive_wall_single_layer_approx_component(
                          plane, exponents, input_data))

    return Element(length=length, beta_x=beta_x, beta_y=beta_y, components=components, name=name, tag=tag,
                   description=description)


def _zlong_round_taper_RW_approx(frequencies: ArrayLike, gamma: float,
                                 layer: Layer, radius_small: float,
                                 radius_large: float, length: float,
                                 step_size: float = 1e-3) -> ArrayLike:
    """
    Function to compute the longitudinal resistive-wall impedance for a
    round taper, integrating the radius-dependent approximated formula
    for a cylindrical structure (see_zlong_round_single_layer_approx above),
    over the length of the taper.
    :param frequencies: the frequencies (array) (in Hz)
    :param gamma: relativistic mass factor
    :param layer: a layer with material properties (only resistivity,
    relaxation time and magnetic susceptibility are taken into account
    at this stage)
    :param radius_small: the smallest radius of the taper (in m)
    :param radius_large: the largest radius of the taper (in m)
    :param length: the total length of the taper (in m)
    :param step_size: the step size (in the radial direction) for the
    integration (in m)
    :return: the longitudinal impedance at these frequencies
    """
    if isscalar(frequencies):
        frequencies = array(frequencies)
    beta = sqrt(1.-1./gamma**2)
    omega = 2*pi*frequencies.reshape((-1, 1))
    k = omega/(beta*c_light)

    rho = layer.dc_resistivity
    tau = layer.resistivity_relaxation_time
    mu1 = 1.+layer.magnetic_susceptibility
    eps1 = 1. - 1j/(eps0*rho*omega*(1.+1j*omega*tau))
    nu = k*sqrt(1.-beta**2*eps1*mu1)

    coef_long = 1j*omega*mu0/(2.*pi*beta**2*gamma**2)

    npts = int(floor(abs(radius_large-radius_small)/step_size)+1)
    radii = linspace(radius_small, radius_large, npts).reshape((1, -1))
    one_array = ones(radii.shape)

    x1 = k.dot(radii)/gamma
    x1sq = x1**2
    x2 = nu.dot(radii)
    zlong = (coef_long.dot(length / float(npts) * one_array) *
             (sp.k0(x1) / sp.i0(x1) - 1. / (x1sq * (1. / 2. + eps1.dot(one_array) *
                                            sp.kve(1, x2) / (x2 * sp.kve(0, x2)))))
            )

    return trapz(zlong, axis=1)


def _zdip_round_taper_RW_approx(frequencies: ArrayLike, gamma: float,
                                layer: Layer, radius_small: float,
                                radius_large: float, length: float,
                                step_size: float = 1e-3) -> ArrayLike:
    """
    Function to compute the transverse dip. resistive-wall impedance for a
    round taper, integrating the radius-dependent approximated formula
    for a cylindrical structure (see_zdip_round_single_layer_approx above),
    over the length of the taper.
    :param frequencies: the frequencies (array) (in Hz)
    :param gamma: relativistic mass factor
    :param layer: a layer with material properties (only resistivity,
    relaxation time and magnetic susceptibility are taken into account
    at this stage)
    :param radius_small: the smallest radius of the taper (in m)
    :param radius_large: the largest radius of the taper (in m)
    :param length: the total length of the taper (in m)
    :param step_size: the step size (in the radial direction) for the
    integration (in m)
    :return: the transverse dipolar impedance at these frequencies
    """
    if isscalar(frequencies):
        frequencies = array(frequencies)
    beta = sqrt(1.-1./gamma**2)
    omega = 2*pi*frequencies.reshape((-1,1))
    k = omega/(beta*c_light)

    rho = layer.dc_resistivity
    tau = layer.resistivity_relaxation_time
    mu1 = 1.+layer.magnetic_susceptibility
    eps1 = 1. - 1j/(eps0*rho*omega*(1.+1j*omega*tau))
    nu = k*sqrt(1.-beta**2*eps1*mu1)

    coef_dip = 1j*k**2*Z0/(4.*pi*beta*gamma**4)

    npts = int(floor(abs(radius_large-radius_small)/step_size)+1)
    radii = linspace(radius_small,radius_large,npts).reshape((1,-1))
    one_array = ones(radii.shape)

    x1 = k.dot(radii)/gamma
    x1sq = x1**2
    x2 = nu.dot(radii)
    zdip = (
            coef_dip.dot(length / float(npts) * one_array) *
            (sp.k1(x1) / sp.i1(x1) + 4 * beta**2 * gamma**2 / (x1sq * (2 + x2 * sp.kve(0, x2) / (mu1 * sp.kve(1, x2)))))
           )

    return trapz(zdip, axis=1)


def create_taper_RW_approx_component(plane: str, exponents: Tuple[int, int, int, int],
                                     input_data: Union[FlatIW2DInput, RoundIW2DInput],
                                     radius_small: float, radius_large: float,
                                     step_size: float = 1e-3) -> Component:
    """
    Creates a single component object modeling a round or flat taper (flatness
    along the horizontal direction, change of half-gap along the vertical one)
    resistive-wall impedance, using the integration of the radius-dependent
    approximated formula for a cylindrical structure (see
    the above functions), over the length of the taper.
    :param plane: the plane the component corresponds to
    :param exponents: four integers corresponding to (source_x, source_y, test_x, test_y) aka (a, b, c, d)
    :param input_data: an IW2D input object (flat or round). If the input
    is of type FlatIW2DInput and symmetric, we apply to the round formula the
    Yokoya factors for an infinitely flat structure (see e.g. K. Yokoya,
    KEK Preprint 92-196 (1993), and Part. Accel. 41 (1993) pp.221-248,
    https://cds.cern.ch/record/248630/files/p221.pdf),
    while for a single plate we use those from A. Burov and V. Danilov,
    PRL 82,11 (1999), https://doi.org/10.1103/PhysRevLett.82.2286. Other
    kinds of asymmetric structure will raise an error.
    If the input is of type RoundIW2DInput, the structure is in principle
    round but the Yokoya factors put in the input will be used.
    Note that the radius or half-gaps in input_data are not used (replaced
    by the scan from radius_small to radius_large, for the integration).
    :param radius_small: the smallest radius of the taper (in m)
    :param radius_large: the largest radius of the taper (in m)
    :param step_size: the step size (in the radial or vertical direction)
    for the integration (in m)
    :return: A component object
    """
    gamma = input_data.relativistic_gamma
    length = input_data.length

    if isinstance(input_data, FlatIW2DInput):
        if len(input_data.top_layers) > 1:
            raise NotImplementedError("Input data can have only one layer")
        yok_long = 1.
        layer = input_data.top_layers[0]
        radius = input_data.top_half_gap
        if input_data.top_bottom_symmetry:
            yok_dipx = pi**2/24.
            yok_dipy = pi**2/12.
            yok_quax = -pi**2/24.
            yok_quay = pi**2/24.
        elif input_data.bottom_half_gap == inf:
            yok_dipx = 0.25
            yok_dipy = 0.25
            yok_quax = -0.25
            yok_quay = 0.25
        else:
            raise NotImplementedError("For asymmetric structures, only the case of a single plate is implemented; "
                                      "hence the bottom half gap must be infinite")
    elif isinstance(input_data, RoundIW2DInput):
        radius = input_data.inner_layer_radius
        if len(input_data.layers) > 1:
            raise NotImplementedError("Input data can have only one layer")
        layer = input_data.layers[0]
        yok_long = input_data.yokoya_factors[0]
        yok_dipx = input_data.yokoya_factors[1]
        yok_dipy = input_data.yokoya_factors[2]
        yok_quax = input_data.yokoya_factors[3]
        yok_quay = input_data.yokoya_factors[4]
    else:
        raise NotImplementedError("Input of type neither FlatIW2DInput nor RoundIW2DInput cannot be handled")

    # Longitudinal impedance
    if plane == 'z' and exponents == (0, 0, 0, 0):
        impedance = lambda f: yok_long*_zlong_round_taper_RW_approx(
                        f, gamma, layer, radius_small, radius_large,
                        length, step_size=step_size)
    # Transverse impedances
    elif plane == 'x' and exponents == (1, 0, 0, 0):
        impedance = lambda f: yok_dipx*_zdip_round_taper_RW_approx(
                        f, gamma, layer, radius_small, radius_large,
                        length, step_size=step_size)
    elif plane == 'y' and exponents == (0, 1, 0, 0):
        impedance = lambda f: yok_dipy*_zdip_round_taper_RW_approx(
                        f, gamma, layer, radius_small, radius_large,
                        length, step_size=step_size)
    elif plane == 'x' and exponents == (0, 0, 1, 0):
        impedance = lambda f: yok_quax*_zdip_round_taper_RW_approx(
                        f, gamma, layer, radius_small, radius_large,
                        length, step_size=step_size)
    elif plane == 'y' and exponents == (0, 0, 0, 1):
        impedance = lambda f: yok_quay*_zdip_round_taper_RW_approx(
                        f, gamma, layer, radius_small, radius_large,
                        length, step_size=step_size)
    else:
        impedance = lambda f: 0

    return Component(impedance=impedance, plane=plane, source_exponents=exponents[:2],
                     test_exponents=exponents[2:])


def create_taper_RW_approx_element(
            input_data: Union[FlatIW2DInput, RoundIW2DInput],
            beta_x: float, beta_y: float,
            radius_small: float, radius_large: float, step_size: float=1e-3,
            component_ids: Sequence[str] = ('zlong', 'zxdip', 'zydip', 'zxqua', 'zyqua'),
            name: str = "", tag: str = "", description: str = "") -> Element:
    """
    Creates an element object modeling a round or flat taper (flatness
    along the horizontal direction, change of half-gap along the vertical one)
    resistive-wall impedance, using the integration of the radius-dependent
    approximated formula for a cylindrical structure (see
    the above functions), over the length of the taper.
    :param input_data: an IW2D input object (flat or round). If the input
    is of type FlatIW2DInput and symmetric, we apply to the round formula the
    Yokoya factors for an infinitely flat structure (see e.g. K. Yokoya,
    KEK Preprint 92-196 (1993), and Part. Accel. 41 (1993) pp.221-248,
    https://cds.cern.ch/record/248630/files/p221.pdf),
    while for a single plate we use those from A. Burov and V. Danilov,
    PRL 82,11 (1999), https://doi.org/10.1103/PhysRevLett.82.2286. Other
    kinds of asymmetric structure will raise an error.
    If the input is of type RoundIW2DInput, the structure is in principle round
    but the Yokoya factors put in the input will be used.
    Note that the radius or half-gaps in input_data are not used (replaced
    by the scan from radius_small to radius_large, for the integration).
    :param beta_x: The beta function in the x-plane at the position of the element
    :param beta_y: The beta function in the y-plane at the position of the element
    :param radius_small: the smallest radius of the taper (in m)
    :param radius_large: the largest radius of the taper (in m)
    :param step_size: the step size (in the radial or vertical direction)
    for the integration (in m)
    :param component_ids: a list of components to be computed
    :param name: A user-specified name for the Element
    :param tag: A string to tag the Element
    :param description: A description for the Element
    :return: An Element object representing the structure
    """
    components = []
    length = input_data.length
    for component_id in component_ids:
        _, plane, exponents = component_names[component_id]
        components.append(create_taper_RW_approx_component(plane=plane, exponents=exponents, input_data=input_data,
                                                           radius_small=radius_small, radius_large=radius_large,
                                                           step_size=step_size))

    return Element(length=length, beta_x=beta_x, beta_y=beta_y, components=components, name=name, tag=tag,
                   description=description)


def create_interpolated_impedance_component(interpolation_frequencies: ArrayLike,
                                            impedance: Optional[Callable] = None,
                                            wake: Optional[Callable] = None, plane: str = '',
                                            source_exponents: Tuple[int, int] = (-1, -1),
                                            test_exponents: Tuple[int, int] = (-1, -1),
                                            name: str = "Unnamed Component",
                                            f_rois: Optional[List[Tuple[float, float]]] = None,
                                            t_rois: Optional[List[Tuple[float, float]]] = None):
    """
    Creates a component in which the impedance function is evaluated directly only on few points and it is interpolated
    everywhere else. This helps when the impedance function is very slow to evaluate.
    :param interpolation_frequencies: the frequencies where the function is evaluated for the interpolation
    :param impedance: A callable function representing the impedance function of the Component. Can be undefined if
    the wake function is defined.
    :param wake: A callable function representing the wake function of the Component. Can be undefined if
    the impedance function is defined.
    :param plane: The plane of the Component, either 'x', 'y' or 'z'. Must be specified for valid initialization
    :param source_exponents: The exponents in the x and y planes experienced by the source particle. Also
    referred to as 'a' and 'b'. Must be specified for valid initialization
    :param test_exponents: The exponents in the x and y planes experienced by the source particle. Also
    referred to as 'a' and 'b'. Must be specified for valid initialization
    :param name: An optional user-specified name of the component
    :param f_rois: a list of frequency regions of interest
    :param t_rois: a list of time regions of interest
    """
    def interpolated_impedance(x):
        return np.interp(x, interpolation_frequencies, impedance(interpolation_frequencies))

    return Component(impedance=interpolated_impedance, wake=wake, plane=plane,
                     source_exponents=source_exponents, test_exponents=test_exponents,
                     name=name, f_rois=f_rois,
                     t_rois=t_rois)



from pywit.devices import create_tesla_cavity_component, shunt_impedance_flat_taper_stupakov_formula

import numpy as np
from pytest import mark

@mark.parametrize("plane, exponents, expected_z",
                  [ ['z', (0, 0, 0, 0), 21382.03077218+1j*26226.15240245],
                    ['x', (1, 0, 0, 0), 84292624.28074735+1j*1.03389207e+08],
                    ['y', (0, 1, 0, 0), 84292624.28074735+1j*1.03389207e+08],
                  ])
def test_create_tesla_cavity_component(plane, exponents, expected_z):
    # using the parameters from SLAC-PUB-9663, "Short-range dipole wakefields in accelerating structures for the NLC"
    period_length=8.75e-3
    a=4.92e-3
    g=6.89e-3
    freq = 1e9
    comp = create_tesla_cavity_component(period_length=period_length, a=a, g=g, plane=plane, exponents=exponents)
    assert np.isclose(np.real(expected_z), np.real(comp.impedance(freq)))
    assert np.isclose(np.imag(expected_z), np.imag(comp.impedance(freq)))


@mark.parametrize("component_id, approximate_integrals, expected_r_shunt",
                  [ ['zlong', False, 171.06327842],
                    ['zlong', True, 171.0632785],
                    ['zxdip', False, 89.89264387],
                    ['zxdip', True, 89.8927562],
                    ['zydip', False, 3047.07197441],
                    ['zydip', True, 3136.96523752],
                    ['zxqua', False, -89.89264417],
                    ['zxqua', True, -89.8927562],
                    ['zyqua', False, 89.89264417],
                    ['zyqua', True, 89.8927562],
                    ['zxcst', False, 0.],
                    ['zycst', True, 0.],
                  ])
def test_shunt_impedance_flat_taper_stupakov_formula(component_id, approximate_integrals, expected_r_shunt):
    length = 97e-3
    delta = 17.6e-3
    half_width = 350e-3
    half_gap = 25e-3

    r_shunt = shunt_impedance_flat_taper_stupakov_formula(half_gap_small=half_gap, half_gap_big=half_gap+delta,
                                                          half_width=half_width, taper_slope=delta / length,
                                                          cutoff_frequency=50e9, component_id=component_id,
                                                          approximate_integrals=approximate_integrals)
    assert np.isclose(r_shunt, expected_r_shunt)



from . import component
from . import devices
from . import element
from . import elements_group
from . import interface
from . import landau_damping
from . import materials
from . import model
from . import parameters
from . import plot
from . import sacherer_formula
from . import utilities
from . import utils


import numpy as np
from scipy import constants as cst

from pywit.sacherer_formula import sacherer_formula
from pywit.utilities import create_resonator_component

# values obtained with the old impedance model
# (https://gitlab.cern.ch/IRIS/IW2D/-/blob/master/PYTHON_codes_and_scripts/Impedance_lib_Python/Impedance.py)
tune_shift_nx_ref = np.array([[-1.89467139e-03 + 2.00708263e-06j, -3.59042292e-05 + 4.25385266e-05j,
                               1.83556168e-03 + 2.00708264e-06j],
                              [-1.89467139e-03 + 2.00708463e-06j, -3.59042230e-05 + 4.25385269e-05j,
                               1.83556167e-03 + 2.00708463e-06j],
                              [-1.89467139e-03 + 2.00708662e-06j, -3.59042167e-05 + 4.25385272e-05j,
                               1.83556167e-03 + 2.00708663e-06j],
                              [-1.89467139e-03 + 2.00708863e-06j, -3.59042104e-05 + 4.25385275e-05j,
                               1.83556167e-03 + 2.00708863e-06j],
                              [-1.89467140e-03 + 2.00709063e-06j, -3.59042042e-05 + 4.25385278e-05j,
                               1.83556167e-03 + 2.00709064e-06j]])

tune_shift_m0_ref = (-3.5904229237583215e-05 + 4.253852659650811e-05j)

z_eff_ref = np.array([[673415.90132594 + 9916238.18255225j,
                       7136258.30545439 + 6023289.34727782j,
                       673415.90381543 + 9916238.18553716j],
                      [673416.56964372 + 9916238.98384426j,
                       7136258.35621533 + 6023288.2941135j,
                       673416.57214035 + 9916238.98683727j],
                      [673417.23974633 + 9916239.78730782j,
                       7136258.40689439 + 6023287.24166201j,
                       673417.24224903 + 9916239.79030961j],
                      [673417.91165383 + 9916240.59313587j,
                       7136258.45756425 + 6023286.19008073j,
                       673417.91416421 + 9916240.59614618j],
                      [673418.58579267 + 9916241.40106015j,
                       7136258.50840891 + 6023285.1389662j,
                       673418.58831139 + 9916241.40407765j]])


def test_sacherer_formula_function():
    impedance_func = create_resonator_component(plane='x', exponents=(1, 0, 0, 0), f_r=0.5e9, q=1, r=1e7).impedance

    def impedance_func_correct(x):
        if np.isscalar(x):
            x = np.array([x])
        ind_p = x >= 0
        ind_n = x < 0
        result = np.zeros_like(x, dtype=complex)
        if len(x[ind_p]) > 0:
            result[ind_p] = impedance_func(x[ind_p])
        if len(x[ind_n]) > 0:
            result[ind_n] = -impedance_func(np.abs(x[ind_n])).conjugate()

        return result

    bunch_intensity = 1.4E11
    bunch_length = 3E-9
    omega_rev = 2.0 * np.pi * cst.c / 26658.8832
    tune = 62.28
    gamma = 6.8E6 / cst.value('proton mass energy equivalent in MeV')
    eta = 0.0003475406157 - 1 / gamma ** 2
    rf_cavities_voltage = 12E6
    beta = np.sqrt(1. - 1. / (gamma ** 2))
    p0 = cst.m_p * beta * gamma * cst.c
    h = 35640
    q_s = np.sqrt(cst.e * rf_cavities_voltage * eta * h / (2 * np.pi * beta * cst.c * p0))
    m_max = 1
    n_bunches = 5
    nx_array = np.linspace(0, n_bunches - 1, n_bunches, dtype=int)

    tune_shift_nx, tune_shift_m0, z_eff = sacherer_formula(qp=10, nx_array=nx_array,
                                                           bunch_intensity=bunch_intensity,
                                                           omegas=q_s * omega_rev, n_bunches=n_bunches,
                                                           omega_rev=omega_rev, tune=tune, gamma=gamma, eta=eta,
                                                           bunch_length_seconds=bunch_length,
                                                           m_max=m_max, impedance_function=impedance_func_correct,
                                                           mode_type='sinusoidal')

    assert np.allclose(tune_shift_nx, tune_shift_nx_ref)
    assert np.isclose(tune_shift_m0, tune_shift_m0_ref)
    assert np.allclose(z_eff, z_eff_ref)


def test_sacherer_formula_table():
    resonator_component = create_resonator_component(plane='x', exponents=(1, 0, 0, 0), f_r=0.5e9, q=1, r=1e7)
    freq_impedance_table, impedance_table = resonator_component.impedance_to_array(10 ** 5, 1e-8, 1e15)
    # breakpoint()
    bunch_intensity = 1.4E11
    bunch_length = 3E-9
    omega_rev = 2.0 * np.pi * cst.c / 26658.8832
    tune = 62.28
    gamma = 6.8E6 / cst.value('proton mass energy equivalent in MeV')
    eta = 0.0003475406157 - 1 / gamma ** 2
    rf_cavities_voltage = 12E6
    beta = np.sqrt(1. - 1. / (gamma ** 2))
    p0 = cst.m_p * beta * gamma * cst.c
    h = 35640
    q_s = np.sqrt(cst.e * rf_cavities_voltage * eta * h / (2 * np.pi * beta * cst.c * p0))
    m_max = 1
    n_bunches = 5
    nx_array = np.linspace(0, n_bunches - 1, n_bunches, dtype=int)

    tune_shift_nx, tune_shift_m0, z_eff = sacherer_formula(qp=10, nx_array=nx_array,
                                                           bunch_intensity=bunch_intensity,
                                                           omegas=q_s * omega_rev, n_bunches=n_bunches,
                                                           omega_rev=omega_rev, tune=tune, gamma=gamma, eta=eta,
                                                           bunch_length_seconds=bunch_length,
                                                           m_max=m_max, impedance_table=impedance_table,
                                                           freq_impedance_table=freq_impedance_table,
                                                           mode_type='sinusoidal')

    assert np.allclose(tune_shift_nx, tune_shift_nx_ref)
    assert np.isclose(tune_shift_m0, tune_shift_m0_ref)
    assert np.allclose(z_eff, z_eff_ref)



from math import sqrt, log

from numpy import vectorize, sin, cos

functions = [vectorize(f) for f in [lambda x: 1/x - log(x) + 1j * (x % 3),
                                    lambda x: x % 2 + x % 4 + 1 + 1j * log(x),
                                    lambda x: x / sqrt(x + x ** 2) + 1j * sqrt(abs(x - 5000)) / 10,
                                    lambda x: 5 * sin(x / 1500) + 1j * 7 * (cos(x / 2000)) + 3]]


def relative_error(a: complex, b: complex) -> float:
    if a == 0 and b == 0:
        return 0
    assert a != 0, "First argument cannot be zero"
    return abs(a - b) / abs(a)



import os

import pytest

from pywit.interface import import_data_iw2d, create_component_from_data, Sampling
from pywit.interface import check_valid_working_directory, get_component_name
from pywit.interface import check_already_computed, get_iw2d_config_value, RoundIW2DInput, FlatIW2DInput, add_iw2d_input_to_database
from pywit.parameters import *
from pywit.materials import layer_from_json_material_library

from pathlib import Path
import glob
from hashlib import sha256
from pytest import raises, mark

import numpy as np

CURR_DIR = Path(__file__).resolve().parent

try:
    import IW2D
    iw2d_installed = True
except ImportError:
    iw2d_installed = False

@mark.parametrize('is_impedance, plane, exponents, expected_comp_name',
                  [[True, 'z', (0, 0, 0, 0), 'zlong'],
                   [False, 'y', (0, 1, 0, 0), 'wydip'],
                   [True, 'x', (0, 0, 1, 0), 'zxqua'],
                   [False, 'y', (0, 0, 0, 0), 'wycst'],
                   ])
def test_get_component_name(is_impedance, plane, exponents, expected_comp_name):
    assert get_component_name(is_impedance, plane, exponents) == expected_comp_name


@mark.parametrize('is_impedance, plane, exponents',
                  [[None, 'l', (0, 0, 0, 0)],
                   [True, 'l', (0, 1, 0, 0)],
                   [False, 'x', ('a', 0, 1, 0)],
                   ])
def test_get_component_name_raise(is_impedance, plane, exponents):
    with raises(ValueError) as error_message:
        get_component_name(is_impedance, plane, exponents)

    assert error_message.value.args[0] == f"({is_impedance},{plane},{exponents}) cannot be found in" \
                                          " the values of component_names dictionary"


@pytest.mark.skipif(not iw2d_installed, reason="IW2D is not installed")
def test_duplicate_component_iw2d_import():
    with raises(AssertionError) as error_message:
        import_data_iw2d(directory=CURR_DIR / "test_data/iw2d/duplicate_components",
                         common_string="WLHC_2layersup_0layersdown6.50mm")

    assert error_message.value.args[0] in ["The wake files 'WlongWLHC_2layersup_0layersdown6.50mm.dat' and "
                                           "'Wlong2WLHC_2layersup_0layersdown6.50mm.dat' both correspond to the "
                                           "z-plane with exponents (0, 0, 0, 0).",
                                           "The wake files 'Wlong2WLHC_2layersup_0layersdown6.50mm.dat' and "
                                           "'WlongWLHC_2layersup_0layersdown6.50mm.dat' both correspond to the "
                                           "z-plane with exponents (0, 0, 0, 0).",
                                           ]


@pytest.mark.skipif(not iw2d_installed, reason="IW2D is not installed")
def test_no_matching_filename_iw2d_import():
    test_dir = CURR_DIR / "test_data/iw2d/valid_directory"

    with raises(AssertionError) as error_message:
        import_data_iw2d(directory=test_dir,
                         common_string="this_string_matches_no_file")

    expected_error_message = f"No files in " \
                             f"'{test_dir.resolve()}'" \
                             f" matched the common string 'this_string_matches_no_file'."

    assert error_message.value.args[0] == expected_error_message


@pytest.mark.skipif(not iw2d_installed, reason="IW2D is not installed")
def test_valid_iw2d_component_import():
    # Normally, the relativistic gamma would be an attribute of a required IW2DInput object, but here it has been
    # hard-coded instead
    relativstic_gamma = 479.605064966
    recipes = import_data_iw2d(directory=CURR_DIR / "test_data/iw2d/valid_directory",
                               common_string="precise")
    for recipe in recipes:
        component = create_component_from_data(*recipe, relativistic_gamma=relativstic_gamma)
        data = recipe[-1]
        x = data[:, 0]
        y_actual = data[:, 1] + (1j * data[:, 2] if data.shape[1] == 3 else 0)
        if recipe[0]:
            np.testing.assert_allclose(y_actual, component.impedance(x), rtol=REL_TOL, atol=ABS_TOL)
        else:
            np.testing.assert_allclose(y_actual, component.wake(x), rtol=REL_TOL, atol=ABS_TOL)


@pytest.fixture
def iw2d_input(request):
    f_params = Sampling(start=1, stop=1e9, scan_type=0, added=(1e2,))
    z_params = Sampling(start=1e-9, stop=1, scan_type=0, added=(1e-6,))
    layers_tung = (layer_from_json_material_library(thickness=np.inf, material_key='W'),)

    if request.param['chamber_type'] == 'round':
        return RoundIW2DInput(machine='test', length=1, relativistic_gamma=7000,
                              calculate_wake=request.param['wake_computation'], f_params=f_params, comment='test',
                              layers=layers_tung, inner_layer_radius=5e-2, yokoya_factors=(1, 1, 1, 1, 1), z_params=z_params)

    if request.param['chamber_type'] == 'flat':
        return FlatIW2DInput(machine='test', length=1, relativistic_gamma=7000,
                             calculate_wake=request.param['wake_computation'], f_params=f_params, comment='test',
                             top_bottom_symmetry=True, top_layers=layers_tung, top_half_gap=5e-2, z_params=z_params)


def _remove_non_empty_directory(directory_path: Path):
    if not os.path.exists(directory_path):
        raise ValueError(f"directory {directory_path} doesn't exist")

    for root, _, files in os.walk(directory_path, topdown=False):
        for name in files:
            os.remove(os.path.join(root, name))

    os.rmdir(directory_path)


list_of_inputs_to_test = [({'chamber_type': 'round', 'wake_computation': False}, ['Zlong', 'Zxdip', 'Zydip', 'Zxquad', 'Zyquad']),
                          ({'chamber_type': 'flat', 'wake_computation': False}, ['Zlong', 'Zxdip', 'Zydip', 'Zxquad', 'Zyquad', 'Zycst']),
                          ({'chamber_type': 'round', 'wake_computation': True}, ['Zlong', 'Zxdip', 'Zydip', 'Zxquad', 'Zyquad', 'Wlong', 'Wxdip', 'Wydip', 'Wxquad', 'Wyquad']),
                          ({'chamber_type': 'flat', 'wake_computation': True}, ['Zlong', 'Zxdip', 'Zydip', 'Zxquad', 'Zyquad', 'Zycst', 'Wlong', 'Wxdip', 'Wydip', 'Wxquad', 'Wyquad', 'Wycst'])]
@pytest.mark.parametrize("iw2d_input, components_to_test", list_of_inputs_to_test, indirect=["iw2d_input"])
@pytest.mark.skipif(not iw2d_installed, reason="IW2D is not installed")
def test_check_already_computed(iw2d_input, components_to_test):
    name = 'test_hash'

    # create the expected directories for the dummy input
    projects_path = Path(get_iw2d_config_value('project_directory'))
    input_hash = sha256(iw2d_input.__str__().encode()).hexdigest()
    directory_level_1 = projects_path.joinpath(input_hash[0:2])
    directory_level_2 = directory_level_1.joinpath(input_hash[2:4])
    working_directory = directory_level_2.joinpath(input_hash[4:])

    if not os.path.exists(directory_level_1):
        os.mkdir(directory_level_1)

    if not os.path.exists(directory_level_2):
        os.mkdir(directory_level_2)

    # check if the directory already existed, otherwise remove it with the content
    if os.path.exists(working_directory):
        [os.remove(old_file) for old_file in glob.glob(f'{working_directory}/*')]
    else:
        os.mkdir(working_directory)

    dummy_string = 'dummy_string'
    for comp in components_to_test:
        with open(f'{working_directory}/{comp}_test.txt', 'w') as f:
            f.write(dummy_string)

    # check that the input is detected in the hashmap
    already_computed, input_hash, working_directory = check_already_computed(iw2d_input, name)
    assert already_computed

    # now we remove one component and verify that check_already_computed gives false
    os.remove(f'{working_directory}/Zlong_test.txt')
    already_computed, input_hash, working_directory = check_already_computed(iw2d_input, name)
    assert not already_computed

    # now we remove the folder and check that check_already_computed gives false
    _remove_non_empty_directory(working_directory)

    already_computed, input_hash, working_directory = check_already_computed(iw2d_input, name)

    assert not already_computed

    # check_already_computed creates working_directory again so we clean it up
    _remove_non_empty_directory(working_directory)


@pytest.mark.parametrize("iw2d_input", [{'chamber_type': 'round', 'wake_computation': False}], indirect=["iw2d_input"])
@pytest.mark.skipif(not iw2d_installed, reason="IW2D is not installed")
def test_add_iw2d_input_to_database(iw2d_input):
    projects_path = Path(get_iw2d_config_value('project_directory'))
    input_hash = sha256(iw2d_input.__str__().encode()).hexdigest()
    directory_level_1 = projects_path.joinpath(input_hash[0:2])
    directory_level_2 = directory_level_1.joinpath(input_hash[2:4])
    working_directory = directory_level_2.joinpath(input_hash[4:])

    add_iw2d_input_to_database(iw2d_input, input_hash, working_directory)

    assert os.path.exists(f"{working_directory}/input.txt")

    _remove_non_empty_directory(working_directory)

    # delete the upper level directories if they are empty
    if not any(os.scandir(directory_level_2)):
        os.rmdir(directory_level_2)

    if not any(os.scandir(directory_level_1)):
        os.rmdir(directory_level_1)


@pytest.mark.skipif(not iw2d_installed, reason="IW2D is not installed")
def test_check_valid_working_directory():
    projects_path = Path(get_iw2d_config_value('project_directory'))
    working_directory = projects_path.joinpath(Path("a/wrong/working_directory"))

    assert not check_valid_working_directory(working_directory=working_directory)



import pytest
import numpy as np

from pywit.landau_damping import dispersion_integral_2d, find_detuning_coeffs_threshold
from pywit.landau_damping import find_detuning_coeffs_threshold_many_tune_shifts


def test_dispersion_integral_2d():

    distribution = 'gaussian'
    tune_shift = -7.3622423693e-05-3.0188372754e-06j
    b_direct = 7.888357197059519e-05
    b_cross = -5.632222163778799e-05

    # reference value obtained with DELPHI
    expected_value = 7153.859519171599-2722.966574677259j

    assert np.isclose(np.real(expected_value), np.real(dispersion_integral_2d(tune_shift=tune_shift,
                                                                              b_direct=b_direct,
                                                                              b_cross=b_cross,
                                                                              distribution=distribution)))
    assert np.isclose(np.imag(expected_value), np.imag(dispersion_integral_2d(tune_shift=tune_shift,
                                                                              b_direct=b_direct,
                                                                              b_cross=b_cross,
                                                                              distribution=distribution)))

    distribution = 'parabolic'

    # reference value obtained with DELPHI
    expected_value = 6649.001778623455-3168.4257879737443j

    assert np.isclose(np.real(expected_value), np.real(dispersion_integral_2d(tune_shift=tune_shift,
                                                                              b_direct=b_direct,
                                                                              b_cross=b_cross,
                                                                              distribution=distribution)))
    assert np.isclose(np.imag(expected_value), np.imag(dispersion_integral_2d(tune_shift=tune_shift,
                                                                              b_direct=b_direct,
                                                                              b_cross=b_cross,
                                                                              distribution=distribution)))


def test_find_detuning_coeffs_threshold():
    # reference values obtained with the old impedance wake model https://gitlab.cern.ch/IRIS/HLLHC_IW_model
    # test positive octupole polarity
    tune_shift = -7.3622423693e-05 - 3.0188372754e-06j
    q_s = 2e-3
    b_direct = 1.980315192200037e-05
    b_cross = -1.4139287608495406e-05
    assert np.isclose(b_direct, find_detuning_coeffs_threshold(tune_shift=tune_shift, q_s=q_s, b_direct_ref=b_direct,
                                                               b_cross_ref=b_cross)[0])
    assert np.isclose(b_cross, find_detuning_coeffs_threshold(tune_shift=tune_shift, q_s=q_s, b_direct_ref=b_direct,
                                                              b_cross_ref=b_cross)[1])
    # test negative octupole polarity
    b_direct = -1.2718161244917965e-05
    b_cross = 9.08066253298482e-06

    assert np.isclose(b_direct, find_detuning_coeffs_threshold(tune_shift=tune_shift, q_s=q_s, b_direct_ref=b_direct,
                                                               b_cross_ref=b_cross)[0])
    assert np.isclose(b_cross, find_detuning_coeffs_threshold(tune_shift=tune_shift, q_s=q_s, b_direct_ref=b_direct,
                                                              b_cross_ref=b_cross)[1])


def test_find_detuning_coeffs_threshold_many_tune_shifts():
    # reference values obtained with the old impedance wake model https://gitlab.cern.ch/IRIS/HLLHC_IW_model
    tune_shift = -7.3622423693e-05 - 3.0188372754e-06j
    q_s = 2e-3
    tune_shifts = [np.nan, tune_shift, 2*tune_shift]

    # test positive octupole polarity
    b_direct = 3.960630384598084e-05
    b_cross = -2.8278575218404585e-05
    assert np.isclose(b_direct, find_detuning_coeffs_threshold_many_tune_shifts(tune_shifts=tune_shifts, q_s=q_s,
                                                                                b_direct_ref=b_direct,
                                                                                b_cross_ref=b_cross)[0])
    assert np.isclose(b_cross, find_detuning_coeffs_threshold_many_tune_shifts(tune_shifts=tune_shifts, q_s=q_s,
                                                                               b_direct_ref=b_direct,
                                                                               b_cross_ref=b_cross)[1])
    # test negative octupole polarity
    b_direct = -2.5436322491034757e-05
    b_cross = 1.816132506682559e-05
    assert np.isclose(b_direct, find_detuning_coeffs_threshold_many_tune_shifts(tune_shifts=tune_shifts, q_s=q_s,
                                                                                b_direct_ref=b_direct,
                                                                                b_cross_ref=b_cross)[0])
    assert np.isclose(b_cross, find_detuning_coeffs_threshold_many_tune_shifts(tune_shifts=tune_shifts, q_s=q_s,
                                                                               b_direct_ref=b_direct,
                                                                               b_cross_ref=b_cross)[1])



from .element import Element

from typing import List, Optional, Tuple


class Model:
    """
    Suggestion for structure of Model class
    """
    def __init__(self, elements: List[Element] = None, lumped_betas: Optional[Tuple[float, float]] = None):
        assert elements, "Model object needs to be initialized with at least one Element"
        if lumped_betas is not None:
            elements = [element.changed_betas(*lumped_betas) for element in elements]

        self.__elements = elements
        self.__lumped_betas = lumped_betas

    @property
    def elements(self):
        return self.__elements

    @property
    def total(self):
        return sum(self.__elements)

    def append_element(self, element: Element):
        if self.__lumped_betas is not None:
            element = element.changed_betas(*self.__lumped_betas)
        self.__elements.append(element)



'''
generate files to mirror in pywit the modules from xwakes.wit
'''

import os

fff = os.listdir('../xwakes/wit')

for nn in fff:
    if not nn.endswith('.py') or nn.startswith('_'):
        continue

    with open(nn, 'w') as fid:
        fid.write(f'from xwakes.wit.{nn.split(".py")[0]} import *\n')


from pywit import materials

from numpy import testing
import numpy as np
from pytest import raises,mark


@mark.parametrize('missing_key',
                  ['dc_resistivity',
                   'resistivity_relaxation_time',
                   're_dielectric_constant',
                   'magnetic_susceptibility',
                   'permeability_relaxation_frequency',
                   ])
def test_layer_from_dict_missing_key(missing_key):
    input_dict = {'dc_resistivity': 1e-5,
                  'resistivity_relaxation_time': 1e-12,
                  're_dielectric_constant': 1,
                  'magnetic_susceptibility': 0,
                  'permeability_relaxation_frequency': np.inf,
                  }
    input_dict.pop(missing_key)
    with raises(AssertionError, match=f"{missing_key} missing from the input dictionary"):
        materials.layer_from_dict(1.,input_dict)


def test_layer_from_dict_two_missing_keys():
    input_dict = {'resistivity_relaxation_time': 1e-12,
                  'magnetic_susceptibility': 0,
                  'permeability_relaxation_frequency': np.inf,
                  }
    with raises(AssertionError, match=f"dc_resistivity, re_dielectric_constant missing from the input dictionary"):
        materials.layer_from_dict(1.,input_dict)


@mark.parametrize("T, B, rho",
                  [ [20, 0.5, 2.6e-10],
                    [20, 8.3, 7.6e-10],
                  ])
def test_copper(T,B,rho):
    RRR = 70.
    copper_layer = materials.copper_at_temperature(1., T, RRR, B)
    testing.assert_allclose(copper_layer.dc_resistivity, rho, atol=1e-11)


@mark.parametrize("material_key, material_function, T, RRR",
                  [ ["Cu", 'copper_at_temperature', 293, 70],
                    ["W",  'tungsten_at_temperature', 300, 70],
                  ])
@mark.parametrize("material_property, tolerance",
                  [ ['dc_resistivity', 2e-9],
                    ['resistivity_relaxation_time', 3e-15], 
                    ['re_dielectric_constant', 0],
                    ['magnetic_susceptibility', 0],
                    ['permeability_relaxation_frequency', 0],
                  ])
def test_materials_at_temperature(material_key, material_function, T,
                                  RRR, material_property, tolerance):
    B = 0.
    layer_at_temperature = getattr(materials,material_function)(1., T, RRR, B)
    layer_from_library = materials.layer_from_json_material_library(1.,material_key)
    
    assert layer_at_temperature.thickness == layer_from_library.thickness == 1.
    testing.assert_allclose(getattr(layer_at_temperature,material_property),
                            getattr(layer_from_library,material_property),
                            atol=tolerance)

